#!/usr/bin/env python3

import os

def build_dihe_folder(fname, type_list, charges):
    # Function to store force field strings, and
    # atomtype-charges for the GauScan2log function
    folder = 'dihedrals'
    if not os.path.exists(folder):
       os.makedirs(folder)
       
    with open(fname , 'r') as file:
        file_content = file.read()
        split_content = file_content.split('!Master function')
        if len(split_content) > 1:
            # Storing the text after encountering the '!Master function'
            text_after_mf = split_content[1].strip()
        f1_path = os.path.join(folder, 'ff_string.txt')
        with open(f1_path,'w') as f1:
            f1.write(text_after_mf)
    
    f2_path = os.path.join(folder, 'type_charge.txt')
    with open(f2_path,'w') as f2:
        for type, charge in zip(type_list, charges):
                f2.write(f'{type}-{charge}\n')
                
    
                
    

def print_GauInp(*args):
    ele_list, type_list, coord, bond_type_list, \
    k_bond_list, bond_length_list, angle_type_list, k_angle_list, \
    angle_length_list, torsion_type_list, v1_list, v2_list, \
    v3_list, phase_list, hybrid_list, charges, formal_chg, multi = args

    header_gjf = """%mem=1GB
%nprocshared=1
%chk=SmartField4gau.chk
#p Amber=(SoftFirst,Print) nosymm geom=nocrowd opt(MaxMicroiterations=2000) Freq

Title

{f_chg} {mult}
"""

    master_function = """
!Master function
NonBon 3 1 0 0 0.000 0.000 0.500 0.000 0.000 -1.2
"""
    filename = 'SmartField4gau.gjf'
    with open(filename, 'w') as file_out:
        file_out.write(header_gjf.format(f_chg = formal_chg, mult=multi))
        for element, type, coordinates, charge in zip(ele_list, type_list, coord, charges):
            formatted_coords = '  '.join(f'{x:.6f}' for x in coordinates)
            file_out.write(f'{element}-{type}-{charge}  {formatted_coords}\n')

        file_out.write(master_function)
        file_out.write('!SMARTFIELD FF\n')
        write_bonds(file_out, bond_type_list, k_bond_list, bond_length_list)
        write_angles(file_out, angle_type_list, k_angle_list, angle_length_list)
        write_torsions(file_out, torsion_type_list, phase_list, v1_list, v2_list, v3_list, hybrid_list)
        file_out.write('\n')


def write_bonds(file, bond_type_list, k_bond_list, bond_length_list):
    file.write('!Bonds\n')
    for bond_type, k_bond, bond_length in zip(bond_type_list, k_bond_list, bond_length_list):
        file.write(f'HrmStr1 {bond_type}  {k_bond:.3f} {bond_length:.3f}\n')


def write_angles(file, angle_type_list, k_angle_list, angle_length_list):
    file.write('!Angles\n')
    for angle_type, k_angle, angle_length in zip(angle_type_list, k_angle_list, angle_length_list):
        file.write(f'HrmBnd1 {angle_type}  {k_angle:.3f} {angle_length:.3f}\n')


def write_torsions(file, torsion_type_list, phase_list, v1_list, v2_list, v3_list, hybrid_list):
    file.write('!Torsions\n')
    for torsion_type, phase, v1, v2, v3, hybrid in zip(torsion_type_list, phase_list, v1_list, v2_list, v3_list, hybrid_list):
        formatted_phase = '  '.join(f'{x}' for x in phase)
        file.write(f'AmbTrs {torsion_type} {formatted_phase} {v1:.2f} {v2:.2f} {v3:.2f} 0. {float(hybrid)}\n')


def print_AmbFrcmod(*args):
    type_list, bond_type_list, k_bond_list, bond_length_list, \
    angle_type_list, k_angle_list, angle_length_list, torsion_type_list, \
    v1_list, v2_list, v3_list, phase_list, hybrid_list = args

    """
    Write an AMBER-like frcmod file:
    type_list: atom type list
    bond_type_list: bond type list
    bond_length_list: bond length list 
    k_bond_list: bond constant list
    angle_type_list: angle type list 
    k_angle_list: angle constant list 
    angle_length_list: degree angle list
    """

    type_list_unique = set(type_list)
    filename = 'SmartField_frcmod.txt'
    with open(filename, 'w') as file_out:
        file_out.write('MASS\n')
        for type_entry in type_list_unique:
            file_out.write(f'{type_entry}\n')
        file_out.write('\n')

        write_bonds_amber(file_out, bond_type_list, k_bond_list, bond_length_list)
        write_angles_amber(file_out, angle_type_list, k_angle_list, angle_length_list)
        write_torsions_amber(file_out, torsion_type_list, phase_list, v1_list, v2_list, v3_list, hybrid_list)


def write_bonds_amber(file, bond_type_list, k_bond_list, bond_length_list):
    file.write('BOND\n')
    for bond_type, k_bond, bond_length in zip(bond_type_list, k_bond_list, bond_length_list):
        file.write(f'{bond_type} {k_bond:.3f} {bond_length:.3f}\n')
    file.write('\n')


def write_angles_amber(file, angle_type_list, k_angle_list, angle_length_list):
    file.write('ANGLE\n')
    for angle_type, k_angle, angle_length in zip(angle_type_list, k_angle_list, angle_length_list):
        file.write(f'{angle_type} {k_angle:.3f} {angle_length:.3f}\n')
    file.write('\n')
        


def write_torsions_amber(file, torsion_type_list, phase_list, v1_list, v2_list, v3_list, hybrid_list):
    file.write('DIHE\n')
    for torsion_type, phase, v1, v2, v3, hybrid in zip(torsion_type_list, phase_list, v1_list, v2_list, v3_list, hybrid_list):
        formatted_phase = '  '.join(f'{x}' for x in phase)
        file.write(f'{torsion_type} {formatted_phase} {v1:.2f} {v2:.2f} {v3:.2f} 0. {hybrid}\n')
    file.write('\n')
    




#!/usr/bin/env python3

import re
import argparse
import pandas as pd

def atoi(text):
    return int(text) if text.isdigit() else text

def natural_keys(text):
    '''
    alist.sort(key=natural_keys) sorts in human order
    http://nedbatchelder.com/blog/200712/human_sorting.html
    (See Toothy's implementation in the comments)
    '''
    return [ atoi(c) for c in re.split(r'(\d+)', text) ]

def build_parser():
    par = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)
    txt = 'MM log files'
    par.add_argument('filelist', type=str, help=txt, nargs='+')
    par.add_argument('-o', type=str, help='name output file', default='output.csv', nargs='?')
    par.add_argument('-t', type=str, default='mm', help='type of log file', nargs='?')
    return par

if __name__ == '__main__':
    PAR = build_parser()
    OPTS = PAR.parse_args()
    file_list = OPTS.filelist

    match_mm = " Energy="
    match_qm = " SCF Done:"
    file_list.sort(key=natural_keys)
    res = []
    for file in file_list:
        # print(file)
        with open(file, 'r') as fname:
             for line in fname:
                 if OPTS.t == 'mm':
                    if line[:8] == match_mm:
                       res.append(line[11:28]) 
                 elif OPTS.t == 'qm':
                     if line[:10] == match_qm:
                       res.append(line[23:40]) 
    df =pd.DataFrame(res)
    fout = OPTS.o
    df.to_csv(fout)
                       





#!/usr/bin/env python3

import numpy as np
import os

def range2(start, end):
     return range(start, end+1)

def flat_list(lis):
    flatList = []
    # Iterate with outer list
    for element in lis:
        if type(element) is list:
            # Check if type is list than iterate through the sublist
            for item in element:
                flatList.append(item)
        else:
            flatList.append(element)
    return flatList
 
def store_any_file(fname):
    """ 
    Reading Any file 
    """
    with open(fname, 'r') as f:
        all_lines = []
        for line in f:
            all_lines.append(line.strip())

    return all_lines

def read_XYZ(all_lines):
    """ 
    Reading CC XYZ from fchk file 
    """
    # with open(fname, 'r') as f:
    #     all_lines = []
    #     for line in f:
    #         all_lines.append(line.strip())

    for s in range(len(all_lines)):                          # Get no of Atoms
        if 'Number of atoms' in all_lines[s]:
            Natom = int(all_lines[s][-10:])  
        
    No_cc = 3* Natom
    nlines = int(No_cc/5.0) + 1
    cc_xyz_list = []
    for s in range(len(all_lines)):                          #  Get CC Coordinates
        if 'Current cartesian coordinates' in all_lines[s]:                              #Reads the Input orientation information
            start = s
            for e in range2(start + 1, start + nlines):
                cc_xyz_list.append(all_lines[e].split() )
    
    cc_xyz_flat = flat_list(cc_xyz_list)

    # Take out Garbage Strings
    if len(cc_xyz_flat) != No_cc:
        diff = abs(len(cc_xyz_flat) - No_cc)
        cc_xyz_flat_mod = cc_xyz_flat[:-diff]
        cc_xyz_1D = np.array(cc_xyz_flat_mod, float)
    else:
        cc_xyz_1D = np.array(cc_xyz_flat, float)

    cc_xyz_arr = np.reshape(cc_xyz_1D, (Natom,3))
    cc_xyz_arr = cc_xyz_arr*0.529177                          # Convert from Bohr to Ang.
    
    return Natom, cc_xyz_arr

def read_RicDim_Grad(all_lines):
    """ 
    Reading Redundant Internal Dimension &
    Gradient from fchk file 
    """

    for s in range(len(all_lines)):                          # Get no of Atoms
        if 'Redundant internal dimensions' in all_lines[s]:
            tmp_list = all_lines[s+1].split()
            ric_list = list(map(int,tmp_list))

    force_list = []
    for s in range(len(all_lines)):                          # Get no of Atoms
        if 'Internal Forces' in all_lines[s]:
            N_force = int(all_lines[s][-10:])
            nlines = int(N_force/5.0) + 1
            start = s
            for e in range2(start + 1, start + nlines):
                force_list.append(all_lines[e].split() )
    
    
    force_flat = flat_list(force_list)

    # Take out Garbage Strings
    if len(force_flat) != N_force:
        diff = abs(len(force_flat) - N_force)
        force_flat_mod = force_flat[:-diff]
        force_1D = np.array(force_flat_mod, float)
    else:
        force_1D = np.array(force_flat,float)

    return ric_list, force_1D


def symmetricize(arr1D):
    ID = np.arange(arr1D.size)
    return arr1D[np.abs(ID - ID[:,None])]


def fill_lower_diag(a):
    n = int(np.sqrt(len(a)*2))+1
    mask = np.tri(n,dtype=bool, k=-1) # or np.arange(n)[:,None] > np.arange(n)
    out = np.zeros((n,n),dtype=int)
    out[mask] = a
    return out

def read_HessRIC(all_lines, ric_list):
    """ 
    Reading Internal Hessian from fchk file:
    all_lines : text file
    ric_list = [RICs, BONDs, ANGLEs, DIHEs]  
    """

    hess_list = []
    for s in range(len(all_lines)):                          # Get no of Atoms
        if 'Internal Force Constants' in all_lines[s]:
            N_hess = int(all_lines[s][-10:])
            nlines = int(N_hess/5.0) + 1
            start = s
            for e in range2(start + 1, start + nlines):
                hess_list.append(all_lines[e].split() )
    
    hess_flat = flat_list(hess_list)

    # Take out Garbage Strings
    if len(hess_flat) != N_hess:
        diff = abs(len(hess_flat)-N_hess)
        # print('diff = ', diff)
        hess_flat_mod = hess_flat[:-diff]
        hess_1D = np.array(hess_flat_mod, float)
    else:
        hess_1D = np.array(hess_flat,float)

    
    # hess_1D_mod = np.append(hess_1D, [0])
    # print(hess_1D_mod)

    # print(len(hess_1D))
    ric_len = ric_list[0]
    # print(ric_len)
    hess_arr = np.zeros((ric_len, ric_len))
    # print(hess_arr.shape, hess_arr.size)
    for i in range(ric_len+1):
        # i_low = int( 0.5 * i * (i - 1) + -1 )           # Adding n(n+1)/2 elements in up/low triangular matrix
        i_low = int( 0.5 * i * (i - 1)  )           # Adding n(n+1)/2 elements in up/low triangular matrix
        # i_up = int( 0.5 * i *  (i + 1) + 1 )
        i_up = int( 0.5 * i *  (i + 1)   )
        # print(i, i_low, i_up, hess_1D[i_low:i_up])
        hess_arr[i-1,0:i] =  hess_1D[i_low: i_up]
        hess_arr[0:i,i-1] =  hess_1D[i_low: i_up]

    return hess_arr

def read_HessXYZ(all_lines, N_atom):
    """ 
    Reading XYZ Hessian from fchk file:
    all_lines : text file
    """

    hess_list = []
    for s in range(len(all_lines)):                          # Get no of Atoms
        if 'Cartesian Force Constants' in all_lines[s]:
            N_hess = int(all_lines[s][-10:])
            nlines = int(N_hess/5.0) + 1
            start = s
            for e in range2(start + 1, start + nlines):
                hess_list.append(all_lines[e].split() )
    
    hess_flat = flat_list(hess_list)

    # Take out Garbage Strings
    if len(hess_flat) != N_hess:
        diff = abs(len(hess_flat)-N_hess)
        # print('diff = ', diff)
        hess_flat_mod = hess_flat[:-diff]
        hess_1D = np.array(hess_flat_mod, float)
    else:
        hess_1D = np.array(hess_flat,float)

    hess_1D = hess_1D * ((627.509391)/(0.529117*0.529117))   # From Hartree/bohr to kcal/mol / angstrom
    # hess_1D_mod = np.append(hess_1D, [0])
    # print(hess_1D_mod)

    # print(len(hess_1D))
    len_hess = 3*N_atom
    hess_XYZ = np.zeros((len_hess, len_hess))
    # print(hess_arr.shape, hess_arr.size)
    for i in range(len_hess + 1):
        # i_low = int( 0.5 * i * (i - 1) + -1 )        # Adding n(n+1)/2 elements in up/low triangular matrix
        i_low = int( 0.5 * i * (i - 1)  )              # Adding n(n+1)/2 elements in up/low triangular matrix
        # i_up = int( 0.5 * i *  (i + 1) + 1 )
        i_up = int( 0.5 * i *  (i + 1)   )
        # print(i, i_low, i_up, hess_1D[i_low:i_up])
        hess_XYZ[i-1,0:i] =  hess_1D[i_low: i_up]
        hess_XYZ[0:i,i-1] =  hess_1D[i_low: i_up]

    return hess_XYZ


def read_NamesTypes(all_lines):
    """ 
    Reading Atom Names from log file:
    """
    ele_list = []
    atype_list = []
    if all_lines is not None and len(all_lines) > 0: 
        for item in all_lines:
                if "-" in item:
                    ele_list.append(item.split("-")[0])
                    atype_list.append(item.split("-")[1])
                else:
                    ele_list.append(item)
                    # print(item.split("-")[0])
    
    # Build atype_list from scratch if is null
    if not atype_list:
        for i, var in enumerate(ele_list):
        # if not atype_list:
            atype_list.append(''.join(f'{var}{i}'))
        #    atype_list[i] = ''.join(f'{ele_list[i]}{i}')
        
    ele_list = flat_list(ele_list)
    atype_list = flat_list(atype_list)

    return ele_list, atype_list
    

def read_Top(all_lines, ric_list):
    """ 
    Reading Topology from RIC:
    """
    tmp_list = []
    ric_tot = ric_list[0]
    for s in range(len(all_lines)):                          # Get no of Atoms
        if 'Name' in all_lines[s]:
            start = s
            # print(all_lines[s])
            for e in range(start+2, start + ric_tot+2):
                tmp_list.append(all_lines[e][8:25].strip())
            break
   

    n_bonds = 0
    n_angles = 0
    n_dihe = 0
    bond_list = []
    angle_list = []
    dihe_list = []
    for i in range(len(tmp_list)):
        if tmp_list[i][0] == 'R':
           n_bonds += 1 
           bond_list.append(tmp_list[i][2 : -1].replace(','," ").split())
        elif tmp_list[i][0] == 'A':
           n_angles += 1 
           angle_list.append(tmp_list[i][2 : -1].replace(','," ").split())
        elif tmp_list[i][0] == 'D':
           n_dihe += 1 
           dihe_list.append(tmp_list[i][2 : -1].replace(','," ").split())

    bond_list = [list(map(int, x)) for x in bond_list]
    angle_list = [list(map(int, x)) for x in angle_list]
    dihe_list = [list(map(int, x)) for x in dihe_list]

    return bond_list, angle_list, dihe_list


def read_CM5(all_lines, N_atoms):
         """ 
         Reading CM5 charges from log file 
         """
         match = 'Hirshfeld charges, spin densities, dipoles, and CM5 charges using IRadAn=      5:'
         match_2 = 'Mulliken charges:'
         text = []

         if match in all_lines:
             for s in range(len(all_lines)):                            
                 if match in all_lines[s]:                              
                     start = s
                     for e in range2(start + 2, start + N_atoms+1):
                         text.append(all_lines[e].split())
                     break
             chg = [ x[7] for x in text ]
         else:                                                       # If CM5 not found, using Mulliken
             for s in range(len(all_lines)):                            
                 if match_2 in all_lines[s]:                            
                     start = s
                     for e in range2(start + 2, start + N_atoms+1):
                         text.append(all_lines[e].split())
                     break
             
             chg = [ x[2] for x in text ]

         # Add '+' to positive charge
         chg_mod = []
         for x in chg:
             if x[0] != '-':
                tmp = '+' + x
             else:
                tmp = x
             chg_mod.append(tmp)

         return chg_mod

def read_AmberParm(path, type_list):
    fname = path + '/amber.prm'
    with open(fname, 'r') as f:
       all_lines = []
       for line in f:
           all_lines.append(line.strip()) 

    Amber_list = []
    for s in range(len(all_lines)):                          #  Get CC Coordinates
        if 'VDW' in all_lines[s]:                              #Reads the Input orientation information
            # start = s
            # for e in range(start + 1, len(all_lines)):
            Amber_list.append(all_lines[s].split() )

    VDW_list = []
    for i in type_list:
        for j in Amber_list:
            if i == j[1]:
                VDW_list.append(j)

    VDW_new = []
    for i in VDW_list:
        if i not in VDW_new:
            VDW_new.append(i)
        
    return VDW_new


def read_mass(all_lines, N_atom, mode):
    """ 
    Reading Masses from fchk file:
    all_lines : text file
    """
    
    # semi_last_column = []
    
    # # Iterate through lines until encountering "Temperature" string
    # for line in lines:
    #     if "Temperature" in line:
    #         break
    #     words = line.split()
    #     if len(words) > 0:
    #         semi_last_column.append(words[-2])
    
    # # Display the semi-last column strings
    # print(semi_last_column)
    
    if mode == 'qm':
        mass_list = []
        for s in range(len(all_lines)):                          # Get no of Atoms
            if 'Real atomic weights' in all_lines[s]:
                N_mass = int(all_lines[s][-10:])
                nlines = int(N_mass/5.0) + 1
                start = s
                for e in range2(start + 1, start + nlines):
                    mass_list.append(all_lines[e].split() )
        
        mass_flat = flat_list(mass_list)
        
    
        # Take out Garbage Strings
        if len(mass_flat) != N_mass:
            diff = abs(len(mass_flat)-N_mass)
            # print('diff = ', diff)
            hess_flat_mod = mass_flat[:-diff]
            mass_1D = np.array(hess_flat_mod, float)
        else:
            mass_1D = np.array(mass_flat,float)
        return mass_1D
    elif mode =='mm':
        mass_col = []
        start_capture = False
        # Iterate through lines between 'Temperature' and 'Molecular mass'
        for line in all_lines:
            if "Molecular mass" in line:
                break
            if start_capture:
                words = line.split()
                if len(words) > 0:
                    mass_col.append(words[-1])
            if "Temperature" in line:
                start_capture = True
        mass_1D = np.asarray(mass_col,float)
        return mass_1D



    

   




#!/usr/bin/env python3

import argparse
import os


def dir_path(string):
    if os.path.isdir(string):
        return string
    else:
        raise NotADirectoryError(string)

def commandline_parser():
    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('-f1','--log_file', help='Gaussian QM log file ')
    
    return parser
 
def store_any_file(fname):
    """ 
    Reading Any file 
    """
    with open(fname, 'r') as f:
        all_lines = []
        for line in f:
            all_lines.append(line.strip())

    return all_lines

def read_Top(all_lines):
    """ 
    Reading Topology from RIC:
    """
    
    count = 0
    match = '-'*80
    # Get the Number of RICs
    for s in range(len(all_lines)):                          
        if all_lines[s] == match:
            start = s
            break
    for e in range(start+1, len(all_lines)):                          
        if all_lines[e] == match:
            break
        count += 1
        
    tmp_list = []
    for s in range(len(all_lines)):                          # Get no of Atoms
        if 'Name' in all_lines[s]:
            start = s
            # print(all_lines[s])
            for e in range(start+2, start + count +2):
                tmp_list.append(all_lines[e][8:25].strip())
            break
   
    n_bonds = 0
    n_angles = 0
    n_dihe = 0
    bond_list = []
    angle_list = []
    dihe_list = []
    for i in range(len(tmp_list)):
        if tmp_list[i][0] == 'R':
           n_bonds += 1 
           bond_list.append(tmp_list[i][2 : -1].replace(','," ").split())
        elif tmp_list[i][0] == 'A':
           n_angles += 1 
           angle_list.append(tmp_list[i][2 : -1].replace(','," ").split())
        elif tmp_list[i][0] == 'D':
           n_dihe += 1 
           dihe_list.append(tmp_list[i][2 : -1].replace(','," ").split())

    bond_list = [list(map(int, x)) for x in bond_list]
    angle_list = [list(map(int, x)) for x in angle_list]
    dihe_list = [list(map(int, x)) for x in dihe_list]

    return bond_list, angle_list, dihe_list

def print_topol(*args):
    bl = args[0]
    al = args[1]
    tl = args[2]
    fname = 'topol.txt'
    with open(fname, 'w') as fout:
        fout.write(f'{len(bl)}\n')
        for i in bl:
            fout.write(f' {i[0]} {i[1]}\n')
        fout.write(f'{len(al)}\n')
        for i in al:
            fout.write(f' {i[0]} {i[1]} {i[2]}\n')
        fout.write(f'{len(tl)}\n')
        for i in tl:
            fout.write(f' {i[0]} {i[1]} {i[2]} {i[3]}\n')
            
            

if __name__ == '__main__':
    parser = commandline_parser()
    opts = parser.parse_args()
    f_qm_log = opts.log_file
    text_qm_log = store_any_file(f_qm_log)
    
    # # Reading in Topology in RIC from log file
    bond_list, angle_list, tors_list = read_Top(text_qm_log)
    print_topol(bond_list, angle_list, tors_list)
    
    
    





# Dictionary of all elements matched with their atomic masses.
elements_dict = {'H' : 1.008,'HE' : 4.003, 'LI' : 6.941, 'BE' : 9.012,\
                 'B' : 10.811, 'C' : 12.011, 'N' : 14.007, 'O' : 15.999,\
                 'F' : 18.998, 'NE' : 20.180, 'NA' : 22.990, 'MG' : 24.305,\
                 'AL' : 26.982, 'SI' : 28.086, 'P' : 30.974, 'S' : 32.066,\
                 'CL' : 35.453, 'AR' : 39.948, 'K' : 39.098, 'CA' : 40.078,\
                 'SC' : 44.956, 'TI' : 47.867, 'V' : 50.942, 'CR' : 51.996,\
                 'MN' : 54.938, 'FE' : 55.845, 'CO' : 58.933, 'NI' : 58.693,\
                 'CU' : 63.546, 'ZN' : 65.38, 'GA' : 69.723, 'GE' : 72.631,\
                 'AS' : 74.922, 'SE' : 78.971, 'BR' : 79.904, 'KR' : 84.798,\
                 'RB' : 84.468, 'SR' : 87.62, 'Y' : 88.906, 'ZR' : 91.224,\
                 'NB' : 92.906, 'MO' : 95.95, 'TC' : 98.907, 'RU' : 101.07,\
                 'RH' : 102.906, 'PD' : 106.42, 'AG' : 107.868, 'CD' : 112.414,\
                 'IN' : 114.818, 'SN' : 118.711, 'SB' : 121.760, 'TE' : 126.7,\
                 'I' : 126.904, 'XE' : 131.294, 'CS' : 132.905, 'BA' : 137.328,\
                 'LA' : 138.905, 'CE' : 140.116, 'PR' : 140.908, 'ND' : 144.243,\
                 'PM' : 144.913, 'SM' : 150.36, 'EU' : 151.964, 'GD' : 157.25,\
                 'TB' : 158.925, 'DY': 162.500, 'HO' : 164.930, 'ER' : 167.259,\
                 'TM' : 168.934, 'YB' : 173.055, 'LU' : 174.967, 'HF' : 178.49,\
                 'TA' : 180.948, 'W' : 183.84, 'RE' : 186.207, 'OS' : 190.23,\
                 'IR' : 192.217, 'PT' : 195.085, 'AU' : 196.967, 'HG' : 200.592,\
                 'TL' : 204.383, 'PB' : 207.2, 'BI' : 208.980, 'PO' : 208.982,\
                 'AT' : 209.987, 'RN' : 222.081, 'FR' : 223.020, 'RA' : 226.025,\
                 'AC' : 227.028, 'TH' : 232.038, 'PA' : 231.036, 'U' : 238.029,\
                 'NP' : 237, 'PU' : 244, 'AM' : 243, 'CM' : 247, 'BK' : 247,\
                 'CT' : 251, 'ES' : 252, 'FM' : 257, 'MD' : 258, 'NO' : 259,\
                 'LR' : 262, 'RF' : 261, 'DB' : 262, 'SG' : 266, 'BH' : 264,\
                 'HS' : 269, 'MT' : 268, 'DS' : 271, 'RG' : 272, 'CN' : 285,\
                 'NH' : 284, 'FL' : 289, 'MC' : 288, 'LV' : 292, 'TS' : 294,\
                 'OG' : 294}

# List of all elements to allow for easy inclusion testing.
elements_list = ['H', 'HE', 'LI', 'BE', 'B', 'C', 'N', 'O', 'F', 'NE', 'NA',\
                 'MG', 'AL', 'SI', 'P', 'S', 'CL', 'AR', 'K', 'CA', 'SC', 'TI',\
                 'V', 'CR', 'MN', 'FE', 'CO', 'NI', 'CU', 'ZN', 'GA', 'GE',\
                 'AS', 'SE', 'BR', 'KR', 'RB', 'SR', 'Y', 'ZR', 'NB', 'MO',\
                 'TC', 'RU', 'RH', 'PD', 'AG', 'CD', 'IN', 'SN', 'SB', 'TE',\
                 'I', 'XE', 'CS', 'BA', 'LA', 'CE', 'PR', 'ND', 'PM', 'SM',\
                 'EU', 'GD', 'TB', 'DY', 'HO', 'ER', 'TM', 'YB', 'LU', 'HF',\
                 'TA', 'W', 'RE', 'OS', 'IR', 'PT', 'AU', 'HG', 'TL', 'PB',\
                 'BI', 'PO', 'AT', 'RN', 'FR', 'RA', 'AC', 'TH', 'PA', 'U',\
                 'NP', 'PU', 'AM', 'CM', 'BK', 'CT', 'ES', 'FM', 'MD', 'NO',\
                 'LR', 'RF', 'DB', 'SG', 'BH', 'HS', 'MT', 'DS', 'RG', 'CN',\
                 'NH', 'FL', 'MC', 'LV', 'TS', 'OG']


print('Atomic Mass Calculator')
print('Developed by: Lukas Richters')
print('2017')
print()
print("Enter the formula for the substance by seperating the atomic symbols\
 from their subscripts. ie. C 6 H 12 O 6.")

# Get the substance's formula from user.
formula = input("Enter a formula or type 'q' to quit: ").upper().split()
print()

# Continue to ask for a new formula until the user inputs a "q" for quit.
while formula != ["Q"]:
    
    # Set the conditons for a new loop.
    atomic_mass_float = 0.0
    invalid_input = False
    coefficient = 1
    
    # Loop through each index in the formula.
    for i, ch in enumerate(formula):
        # If the character is an element, check if the next character in the
        # formula is an integer.
        if ch in elements_list:
            element_mass = elements_dict.get(ch)
            # If the next character is an integer, multiply the element's mass 
            # by that integer and add it to the running sum.
            if formula[i + 1].isdigit() == True:
                atomic_mass_float += element_mass * int(formula[i + 1])
            # If not, just add that element's mass to the sum.
            else:
                atomic_mass_float += element_mass
        # If the charcter is an integer
        elif ch.isdigit() == True:
            # If the first index is an integer, assume that that is a
            # coefficient for the formula and multiply the formula by that
            # number
            if i == 0:
                coefficient = int(ch)
            # Make sure the previous index wasn't an integer as well
            # (the integer needs an element to multiply), if it was, let the
            # program know that there is an error.
            elif formula[i - 1].isalpha() == False:
                invalid_input = True
            else:
                pass
        # If the only character in the formula is an integer, let the program
        # know that there is an error.
        elif ch.isdigit() == True and len(formula) == 1:
            invalid_input = True
        # If a character is not an element or an integer, let the program know
        # that there is an error.
        else:
            invalid_input = True
    
    # Muliply the entire atomic mass by the coefficient.
    atomic_mass_float *= coefficient
    
    # If there was an error in the program, display the appropriate error.
    if invalid_input == True:
        # Error for if there is only an integer entered.
        if ch.isdigit() == True and len(formula) == 1:
            print("You must enter at least one element for every integer\
 subscript.")
            print()
        # Error for if there is a float or non-elemnt entered.
        else:
            print("Please enter only the atomic symbols of elements or an integer\
 subscript.")
            print()
    
    # If there was no error, then print the claculated atomic mass
    else:
        print("Atomic Mass: ", round(atomic_mass_float, 3))
        print()
    
    # Ask for another formula.
    formula = input("Enter a formula or type 'q' to quit: ").upper().split()
    print()        
